<?php declare(strict_types=1);

namespace Infomaker\Everyware\Twig;

use Exception;
use Infomaker\Everyware\Support\NewRelicLog;
use Infomaker\Everyware\Twig\Extension\ProjectExtension;
use Twig\Environment;
use Twig\Error\LoaderError;
use Twig\Extension\DebugExtension;
use Twig\Extension\ExtensionInterface;
use Twig\Loader\FilesystemLoader;
use Twig\NodeVisitor\NodeVisitorInterface;
use Twig\RuntimeLoader\RuntimeLoaderInterface;
use Twig\TokenParser\TokenParserInterface;
use Twig\TwigFilter;
use Twig\TwigFunction;
use Twig\TwigTest;

class ViewSetup
{
    /**
     * @var string
     */
    private const CACHE_PATH = '/var/www/twig-compile';

    /**
     * @var self
     */
    private static $instance;

    /**
     * @var Environment
     */
    private $environment;

    /**
     * Contains all registered template directories
     * @var array
     */
    private $registeredDirs = [];

    private $functions = [];

    public function __construct(Environment $environment)
    {
        $this->environment = $environment;
    }

    /**
     * Create new instance if none exist
     *
     * @return ViewSetup
     */
    public static function getInstance(): ViewSetup
    {
        if (static::$instance === null) {
            $debug = is_dev();

            $environment = new Environment(new FilesystemLoader(), [
                'cache' => is_dev() ? false : static::CACHE_PATH,
                'debug' => $debug
            ]);

            $deprecationHandler = new DeprecationHandler($environment);

            // Set up the default Extensions
            $environment->addExtension(new ProjectExtension());

            if ($debug) {
                $environment->addExtension(new DebugExtension());
                $deprecationHandler->searchAndLog();
            } else {
                $deprecationHandler->listen();
            }

            static::$instance = new static($environment);
        }

        return static::$instance;
    }

    public function init(array $options = []): self
    {
         @trigger_error(sprintf(__CLASS__ . '::init is deprecated as of infomaker/everyware-twig v1.0.0.'), E_USER_DEPRECATED);

        return $this;
    }

    /**
     * @param string $namespace
     * @param string $path
     *
     * @return ViewSetup
     */
    public function registerTwigFolder(string $namespace, string $path): self
    {
        // Store namespaces for registration
        if ( ! isset($this->registeredDirs[$namespace])) {
            $this->registeredDirs[$namespace] = [];
        }

        // Don't register paths twice under same namespace
        if (in_array($path, $this->registeredDirs[$namespace], true)) {
            return $this;
        }

        if (is_dir($path)) {
            try {
                $this->addLoaderPath($path, $namespace);
                $this->registeredDirs[$namespace][] = $path;
            } catch (Exception $e) {
                NewRelicLog::logException($e);
            }
        }

        return $this;
    }

    /**
     * Render function of EveryTwig, loads a specific template and works with cached compiled templates.
     *
     * @param string $template
     * @param array  $context
     *
     * @return void
     */
    public function render(string $template, array $context = []): void
    {
        try {
            $this->environment->display($template, $context);
        } catch (Exception $e) {
            NewRelicLog::logException($e);
        }
    }

    /**
     * Return the view as rendered HTML by passed template
     *
     * @param string $template
     * @param array  $context
     *
     * @return string
     */
    public function view(string $template, array $context = []): string
    {
        try {
            return $this->environment->render($template, $context);
        } catch (Exception $e) {
            NewRelicLog::logException($e);
        }

        return '';
    }

    public function addExtension(ExtensionInterface $extension): void
    {
        try {
            $this->environment->addExtension($extension);
        } catch (\LogicException $e) {
            NewRelicLog::logException($e);
        }
    }

    public function addFilter(string $name, $callable = null, array $options = []): self
    {
        try {
            $this->environment->addFilter(new TwigFilter($name, $callable, $options));
        } catch (\LogicException $e) {
            NewRelicLog::logException($e);
        }

        return $this;
    }

    public function addFunction(string $name, $callable = null, array $options = []): self
    {
        try {
            $this->environment->addFunction(new TwigFunction($name, $callable, $options));
        } catch (\LogicException $e) {
            NewRelicLog::logException($e);
        }

        return $this;
    }

    public function addGlobal(string $key, $value): self
    {
        try {
            $this->environment->addGlobal($key, $value);
        } catch (\LogicException $e) {
            NewRelicLog::logException($e);
        }

        return $this;
    }

    public function addNodeVisitor(NodeVisitorInterface $visitor): void
    {
        try {
            $this->environment->addNodeVisitor($visitor);
        } catch (\LogicException $e) {
            NewRelicLog::logException($e);
        }
    }

    public function addRuntimeLoader(RuntimeLoaderInterface $loader): void
    {
        $this->environment->addRuntimeLoader($loader);
    }

    public function addTest(string $name, $callable = null, array $options = []): self
    {
        try {
            $this->environment->addTest(new TwigTest($name, $callable, $options));
        } catch (\LogicException $e) {
            NewRelicLog::logException($e);
        }

        return $this;
    }

    public function addTokenParser(TokenParserInterface $parser): void
    {
        try {
            $this->environment->addTokenParser($parser);
        } catch (\LogicException $e) {
            NewRelicLog::logException($e);
        }
    }

    public function addWpFunction(string $functionName, string $name = null): self
    {
        $name = ( ! empty($name) && $name !== null) ? $name : $functionName;
        $this->addFunction($name, $functionName);

        return $this;
    }

    public function hasExtension(string $class): bool
    {
        return $this->environment->hasExtension($class);
    }

    /**
     * @param $path
     * @param $namespace
     *
     * @throws LoaderError
     */
    protected function addLoaderPath($path, $namespace): void
    {
        $loader = $this->environment->getLoader();

        if ( ! $loader instanceof FilesystemLoader) {
            throw new LoaderError(sprintf('Could not add path "%s" to Loader ("%s") .', $path, get_class($loader)));
        }

        $loader->addPath($path, $namespace);
        $this->environment->setLoader($loader);
    }
}
